import os
import inspect

from functools import wraps
from types import FunctionType
from collections import MutableSequence, MutableSet

from toolkit.async_context import contextmanager
from apistellar.helper import proxy, get_callargs


class ConnectionManager(object):
    proxy_driver_names = None
    meta_name = "__driver_meta"

    @staticmethod
    def debug_callback():
        return os.getenv("UNIT_TEST_MODE", "").lower() == "true"

    def __init__(self, debug_callback=None, proxy_driver_names: tuple=None,
                 asyncable=False, asyncgen=False):
        self.asyncable = asyncable
        self.asyncgen = asyncgen

        if debug_callback:
            self.debug_callback = debug_callback
        if proxy_driver_names is not None:
            assert isinstance(proxy_driver_names,
                              (tuple, MutableSequence, MutableSet)), \
                "proxy_driver_names TypeError, " \
                "tuple, MutableSequence, MutableSet need!"
            self.proxy_driver_names = proxy_driver_names

    @staticmethod
    def get_generator(func, self_or_cls, need_proxy, *args, **kwargs):
        callargs = get_callargs(func, self_or_cls, *args, **kwargs)
        callargs.pop("cls", None)
        # 将need_proxy代理到self_or_cls中
        self_or_cls = proxy(self_or_cls, need_proxy, "_need_proxy")
        return self_or_cls, self_or_cls.get_store(self_or_cls, **callargs)

    def __call__(self, *args, debug_callback=None, proxy_driver_names=None,
                 asyncable=False, asyncgen=False):
        """
        返回连接管理下的方法
        :param func:
        :param proxy_driver_names: 可以被代理的驱动名称
        :param asyncable: 有些方法可能是同步的，但是通过返回future来变成异步的
        :param asyncgen: 有些方法可能返回异步生成器。
        :return:
        """
        if not args:
            return self.__class__(
                debug_callback, proxy_driver_names, asyncable, asyncgen)

        func = args[0]

        def need_proxy(driver_name):
            """
            DriverMixin的子类在实现get_store时，
            可以调用need_proxy来决定是否为实例添加一个driver， 如：
            ```
            with super().get_store(self_or_cls, **callargs) as self_or_cls:
                if self_or_cls._need_proxy("store"):
                    self_or_cls = proxy(self_or_cls, prop_name="store", prop=...)
                try:
                    yield self_or_cls
                finally:
                    conn.commit()

            ```
            :param driver_name: 添加的driver在实例中的属性名字。
            :return:
            """
            if self.proxy_driver_names is None:
                return True
            return driver_name in self.proxy_driver_names

        if self.asyncable or inspect.iscoroutinefunction(func):
            @wraps(func)
            async def inner(self_or_cls, *args, **kwargs):
                # 将self.debug_callback()写在里面的原因是因为可运行时改变是否debug
                if self.debug_callback():
                    return await func(self_or_cls, *args, **kwargs)
                self_or_cls, gen = self.get_generator(
                    func, self_or_cls, need_proxy, *args, **kwargs)

                async with gen as proxy_instance:
                    return await func(proxy_instance, *args, **kwargs)
        elif self.asyncgen or inspect.isasyncgenfunction(func):
            @wraps(func)
            async def inner(self_or_cls, *args, **kwargs):
                if self.debug_callback():
                    async for i in func(self_or_cls, *args, **kwargs):
                        yield i
                else:
                    self_or_cls, gen = self.get_generator(
                        func, self_or_cls, need_proxy, *args, **kwargs)

                    async with gen as proxy_instance:
                        async for i in func(proxy_instance, *args, **kwargs):
                            yield i
        else:
            @wraps(func)
            def inner(self_or_cls, *args, **kwargs):
                if self.debug_callback():
                    return func(self_or_cls, *args, **kwargs)

                self_or_cls, gen = self.get_generator(
                    func, self_or_cls, need_proxy, *args, **kwargs)

                with gen as proxy_instance:
                    return func(proxy_instance, *args, **kwargs)
        return inner


conn_manager = ConnectionManager()


def conn_meta_add(meta_name):
    def outer(meta_val):
        def inner(func):
            driver_meta = getattr(func, ConnectionManager.meta_name, dict())
            driver_meta[meta_name] = meta_val
            func.__driver_meta = driver_meta
            return func
        return inner
    return outer


# 对于使用PersistentMeta创建的类，conn_manager需要的参数通过这些装饰器来指定。
conn_asyncable = conn_meta_add("asyncable")(True)
conn_asyncgen = conn_meta_add("asyncgen")(True)
conn_proxy_driver_names = conn_meta_add("proxy_driver_names")
conn_debug = conn_meta_add("debug_callback")
ignore_callback = conn_debug(lambda: True)

conn_ignore = ignore_callback


class DriverMixin(object):
    """
    配合conn_manager用来控制数据库访问。
    DriverMixin实现需要注意以下三点：
    1 所有Mixin都继承于DriverMixin(或其子类)，使用super调用父类的get_store方法，
    2 get_store需要被contextmanager装饰，contextmanager(非内置)来自于toolkit.async_context。
    3 Mixin支持将get_store实现成异步的。
    """

    @classmethod
    @contextmanager
    def get_store(cls, self_or_cls, **callargs):
        """
        子类需要通过super调用父类的get_store方法
        :param self_or_cls: 调用类方法和实例方法时的cls或self
        :param callargs: 方法调用时参数表
        :return: 返回proxy对象
        """
        yield self_or_cls


class PersistentMeta(type):
    """
    为实例方法和类方法增加conn_manager装饰器
    """
    def __new__(mcs, name, bases, attrs):
        for attr_name in attrs.keys():
            func = attrs[attr_name]
            # 类方法
            if hasattr(func, "__func__"):
                driver_meta = getattr(func.__func__, ConnectionManager.meta_name, None)
            else:
                driver_meta = getattr(func, ConnectionManager.meta_name, None)
            if driver_meta:
                cm = conn_manager(**driver_meta)
            else:
                cm = conn_manager
            # 去掉魔术方法和私有方法
            if isinstance(func, FunctionType) and not func.__name__.startswith(
                    "__"):
                attrs[attr_name] = cm(func)
            if isinstance(func, classmethod):
                func = func.__func__
                if not func.__name__.startswith("__"):
                    attrs[attr_name] = classmethod(cm(func))

        return super(PersistentMeta, mcs).__new__(mcs, name, bases, attrs)



# 过时
from .metas import ModelFactoryMeta
from .components import Component


class ModelFactory(Component, metaclass=ModelFactoryMeta):
    model = object

    async def product(self, *args, **kwargs):
        pass



import re
import os
import sys
import string

from collections import OrderedDict
from importlib import import_module
from os import makedirs, sep, listdir, getcwd
from os.path import join, exists, abspath, dirname, isdir, basename

from toolkit import load
from apistellar.types import validators
from apistellar.document import DocPainter
from toolkit.tools.markdown_helper import MarkDownRender


__all__ = ["Task", "Project", "Service", "Model", "Solo", "Document", "Rpc"]


class Task(object):
    def __init__(self):
        self.template = None
        self.kwargs = {}

    def create(self, env, **kwargs):
        task = kwargs.pop("task")
        names = kwargs.pop("name", [])
        self.kwargs.update(kwargs)

        for name in names:
            self.enrich_kwargs(self.validate_name(name))
            makedirs(self.kwargs["dirname"], exist_ok=True)
            self.copytree(env, task)
        print("、".join(names), "已创建。")

    def validate_name(self, name):
        return name

    def enrich_kwargs(self, name):
        self.kwargs["name"] = name
        self.kwargs["dirname"] = name

    @classmethod
    def enrich_parser(cls, sub_parser):
        pass

    def copytree(self, env, task, dest_path=None):
        if dest_path is None:
            dest_path = self.kwargs["dirname"]
        copy_path = join(self.template, task)

        for file in listdir(copy_path):
            file = join(copy_path, file)

            if file.count("__pycache__"):
                continue

            if isdir(file):
                dir_name = abspath(join(dest_path, self.render_path_name(file)))
                makedirs(dir_name, exist_ok=True)
                self.copytree(env, file, dir_name)
            else:
                template = env.get_template(file.replace(self.template, ""))
                file = self.render_path_name(file)
                filename = abspath(join(dest_path, file).replace(".tmpl", ""))
                if exists(filename) and \
                        input(f"{filename}已存在，是否覆盖y/n?") not in ["y", "yes"]:
                    continue

                with open(filename, "w") as f:
                    f.write(template.render(**self.kwargs))
                    f.write("\n")

    def render_path_name(self, path):
        return string.Template(basename(path)).substitute(**self.kwargs)


class Project(Task):
    """
    项目
    """
    def validate_name(self, name):
        assert re.search(r'^[_a-zA-Z]\w*$', name), \
            '项目名称只能包含数字、字母、下划线并以字母开头！'
        try:
            mod = import_module(name)
        except ImportError:
            mod = None
        assert mod is None, f'模块 {name} 已存在！'
        return name

    def enrich_kwargs(self, name):
        super(Project, self).enrich_kwargs(name)
        self.kwargs["project_dir"] = join(getcwd(), name)

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs=1, help="项目名称")


class ModuleTask(Task):

    def validate_name(self, name):
        words = re.findall(r"([A-Za-z0-9]+)", name)
        assert words, f"name: {name} invalid!"
        assert words[0][0].isalpha(), f"name: {name} start with number!"
        return words

    def enrich_kwargs(self, words):
        class_name = "".join(word.capitalize() for word in words)
        module_name = "_".join(word.lower() for word in words)
        self.kwargs["class_name"] = class_name
        self.kwargs["module_name"] = module_name
        self.kwargs["dirname"] = module_name


class Service(ModuleTask):
    """
    服务
    """

    def enrich_kwargs(self, words):
        super().enrich_kwargs(words)
        father = None

        if exists("__init__.py"):
            regex = re.compile(r"class\s+(\w*?Controller)\(\w*Controller\):")
            mth = regex.search(open("__init__.py").read())
            if mth:
                father = mth.group(1)
        self.kwargs["father"] = father or "Controller"

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs="+", help="服务模块名称")


class Drivable(Task):

    def enrich_kwargs(self, name):
        super().enrich_kwargs(name)
        driver = self.kwargs.get("driver")
        if driver and driver.count(":") == 1:
            module, cls = driver.split(":")
        else:
            module, cls = None, None
        self.kwargs["module"] = module
        self.kwargs["cls"] = cls

    @classmethod
    def enrich_parser(cls, sub_parser):
        super().enrich_parser(sub_parser)
        sub_parser.add_argument(
            "-d", "--driver", help="驱动 eg：pymongo:MongoClient")


class Model(ModuleTask):
    """
    模型层
    """
    @staticmethod
    def to_list(val):
        """用来兼容接口"""
        return [val]

    def enrich_kwargs(self, words):
        super().enrich_kwargs(words)
        fields = [f.split(":", 1) for f in self.kwargs.get("fields", [])]
        types = dict()

        for p in dir(validators):
            obj = getattr(validators, p)
            if isinstance(obj, type) and \
                    issubclass(obj, validators.Validator):
                types[obj.__name__.lower()] = obj.__name__

        new_fields = []
        for k, v in fields[:]:
            new_fields.append((k, types.get(v.lower(), "")))
        self.kwargs["fields"] = new_fields
        self.kwargs["dirname"] = self.kwargs.get("path") or self.kwargs["dirname"]

    @classmethod
    def enrich_parser(cls, sub_parser):
        super().enrich_parser(sub_parser)
        sub_parser.add_argument(
            "-n", "--name", required=True, type=cls.to_list, help="models名称")
        sub_parser.add_argument(
            "-p", "--path", help="所属服务路径 eg: article/comment")
        # sub_parser.add_argument(
        #     "-a", "--async", action="store_true", help="是否拥有异步获取属性的能力")
        sub_parser.add_argument("fields", nargs="*", help="字段 eg: id:integer")


class Solo(ModuleTask):
    """
    独立任务程序
    """
    def enrich_kwargs(self, words):
        super().enrich_kwargs(words)
        project_name = basename(abspath(getcwd()))
        self.kwargs["project_name"] = project_name
        self.kwargs["dirname"] = join(project_name, self.kwargs["dirname"])
        self.kwargs["back_trace"] = ".." + sep + ".." + sep

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs="+", help="独立任务程序名称")


class APIRenderTask(Task):
    """
    通过API生成文档，RPC调用等
    """
    @staticmethod
    def _cwd(module):
        """
        切换到模块包所在目录并返回
        :param module:
        :return:
        """
        if module:
            current_dir = dirname(load(module).__file__)
            sys.modules.pop(module, None)
        else:
            current_dir = "."
        os.chdir(current_dir)
        return current_dir

    @staticmethod
    def _ignore_input():
        """
        不再提醒是否覆盖
        :return:
        """
        global input

        def _input(prompt):
            return "y"
        input = _input

    def create(self, env, **kwargs):
        """
        创建一个模板实例
        :param env:
        :param kwargs: 解析命令行获取的参数
        :return:
        """
        name = kwargs.pop("name", [])[0]
        self._ignore_input()
        base_dir_name = abspath(join(os.getcwd(), kwargs.pop("location"), name))
        cwd = self._cwd(kwargs["module"])
        indices = OrderedDict()

        for parent, doc in DocPainter(cwd, kwargs["parser"]).paint().items():
            self.kwargs = self._enrich_kwargs(doc, base_dir_name, kwargs)
            self._enrich_indices(doc, indices)
            makedirs(self.kwargs["dirname"], exist_ok=True)
            self.copytree(env, kwargs["task"])

        self._finalize(indices, name, base_dir_name)

    @classmethod
    def _enrich_kwargs(cls, doc, base_dir_name, kwargs):
        doc["dirname"] = base_dir_name
        doc.setdefault("enumerate", enumerate)
        doc.setdefault("bool", bool)
        doc.setdefault("len", len)
        doc.setdefault("map", map)
        return doc

    @staticmethod
    def _enrich_indices(doc: dict, indices: dict):
        """
        在渲染完多个模块时，为能生成一个索引页(或`__init__.py`)保留一些数据
        :param doc:
        :param indices:
        """

    @classmethod
    def _finalize(cls, indices: dict, name: str, base_dir_name: str):
        """
        使indices等数据生成一个索引页或执行其它收尾工作
        :param indices:
        :param name:
        :param base_dir_name:
        :return:
        """

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs=1, help="输出名称")
        sub_parser.add_argument("-m", "--module", help="模块地址", required=True)
        sub_parser.add_argument("-l", "--location", help="输出地址", default=".")
        sub_parser.add_argument(
            "-p", "--parser", help="parser模块地址",
            default="apistellar.document.parser.RstDocParser")


class Document(APIRenderTask):

    @staticmethod
    def _enrich_indices(doc, indices):
        """
        生成index.md所需要的数据
        :param doc:
        :param indices:
        :return:
        """
        fn = join(doc["file_path"], doc["doc_name"] + ".md.html")
        indices[fn] = doc["doc_name"]

    @classmethod
    def _enrich_kwargs(cls, doc, base_dir_name, kwargs):
        """
        为模板增加一个迭代器用来迭代接口
        :param doc:
        :param base_dir_name:
        :param kwargs:
        :return:
        """
        doc = super()._enrich_kwargs(doc, base_dir_name, kwargs)
        doc.setdefault("iter", cls.iter_interface)
        return doc

    @staticmethod
    def iter_interface(interface):
        """
        迭代接口里面的参数
        :param interface:
        :return:
        """
        if "params" in interface:
            yield ("查询参数", interface["params"])

        if "path_params" in interface:
            yield ("路径参数", interface["path_params"])

        if "form_params" in interface:
            yield ("表单参数", interface["form_params"])

    @classmethod
    def _finalize(cls, indices, name, base_dir_name):
        """
        渲染markdown文档成html，并添加一个目录，同时打开文档
        :param indices:
        :param name:
        :param base_dir_name:
        :return:
        """
        os.makedirs(base_dir_name, exist_ok=True)
        with open(os.path.join(base_dir_name, "index.md"), "w") as f:
            cls._write_index(f, indices, name)
            print(f"{name}已创建。")
            with MarkDownRender("github-markdown.css", base_dir_name) as mkr:
                output = mkr.render()

            if output:
                try:
                    os.system(f"open {output}")
                except Exception:
                    print(f"打开{output}失败! ")
            else:
                print("未找到可打开的文档！")
            cls._write_index(f, indices, name, link_html=False)

    @staticmethod
    def _write_index(file, indices, name, link_html=True):
        """
        向目录中写入内容。
        :param file: 目录文件
        :param indices: 目录索引
        :param name: 文档名称
        :param link_html: 目录markdown文件链接到的索引是否是html文件，否则是md文件,
                          之所以以这两遍，是因为第一遍写入html索引连接，用于生成html文档
                          第二遍写入md索引，用于生成md文档，方便在github中查看
        :return:
        """
        file.seek(0)
        file.truncate()
        file.write(f"# {name}\n\n")
        for index, (key, val) in enumerate(indices.items()):
            if not link_html:
                key = key.replace(".md.html", ".md")
            file.write(f"{index + 1}. [{val}]({key})\n")
        file.flush()


class Rpc(APIRenderTask):

    @staticmethod
    def _enrich_indices(doc, indices):
        """
        生成__init__.py所需要的数据
        :param doc:
        :param indices:
        :return:
        """
        indices[doc["controller"].capitalize()] = doc["controller"]

    @classmethod
    def _enrich_kwargs(cls, doc, base_dir_name, kwargs):
        """
        增加一些渲染参数
        :param doc:
        :param base_dir_name:
        :param kwargs:
        :return:
        """
        doc = super()._enrich_kwargs(doc, base_dir_name, kwargs)
        doc["dirname"] = base_dir_name
        base = kwargs["base"]
        assert base.count(":") == 1, f"Invalid format: {base}"
        doc["base_module"], doc["base_class"] = base.split(":")
        doc.setdefault("agg", cls._agg_interface_aiohttp)
        doc.setdefault("str", str)
        if kwargs["conn_timeout"]:
            doc.setdefault(
                "conn_timeout", ", conn_timeout=%d" % kwargs["conn_timeout"])
        if kwargs["read_timeout"]:
            doc.setdefault(
                "read_timeout", ", read_timeout=%d" % kwargs["read_timeout"])
        return doc

    @staticmethod
    def _agg_interface_aiohttp(interface):
        """
        将一个接口中的数据组合成构建方法所需要的部分结构
        :param interface:
        :return:
        """
        args_def = ""
        body_def = ""
        call_args_def = ""
        have_path_param = ""
        if "path_params" in interface:
            args_def += f"path_params: dict, "
            have_path_param = ", have_path_param=True"

        if "json_params" in interface:
            args_def += f"json: dict, "
            call_args_def += ", json=json"

        if "form_params" in interface:
            args_def += f"form_fields: typing.List[dict], "
            form_body = """        data = FormData()
        for meta in form_fields:
            data.add_field(meta["name"],
                           meta["value"],
                           filename=meta.get("filename"),
                           content_type=meta.get("content_type"))\n"""
            call_args_def += ", data=data"
            body_def += form_body

        if "params" in interface:
            params = list()

            for name, param in interface["params"].items():
                args_def += f"{name}: {param['type']}"
                params.append(name)
                # 服务端有默认参数，所以客户端不需要主动传参，默认为None
                if "default" in param:
                    args_def += "=None"
                args_def += ", "

            if params:
                pair = ""
                for name in params:
                    pair += f'        if {name} is not None:' \
                            f'\n            params["{name}"] = {name}\n'

                body_def += "        params = dict()\n%s" % pair
            call_args_def += ", params=params"

        args_def += "cookies: dict=None"

        if "return_wrapped" in interface:
            method = "json"
            success_key_name = interface["return_wrapped"]["success_key_name"]
            success_code = interface["return_wrapped"]["success_code"]
        else:
            success_code = None
            success_key_name = None

            if "return_class" in interface:
                if interface["return_class"] in (bytes, str):
                    method = "read"
                else:
                    method = "json"
            else:
                method = "read"

        if method == "read" or success_code is None:
            error_check = ""
        else:
            error_check = ', error_check=lambda x: x["code"] != %s' % success_code

        if success_key_name is None:
            success_key_name = ""
        else:
            success_key_name = f', "{success_key_name}"'

        return args_def, body_def, call_args_def, method, \
               error_check, success_key_name, have_path_param

    @classmethod
    def _finalize(cls, indices, name, base_dir_name):
        """
        生成一个__init__.py，将所有模块中的接口结合起来
        :param indices:
        :param name:
        :param base_dir_name:
        :return:
        """
        os.makedirs(base_dir_name, exist_ok=True)
        tmpl = f"class {name.capitalize()}(%s):\n    pass"

        with open(os.path.join(base_dir_name, "__init__.py"), "w") as f:
            f.write(f"# {name} API\n")
            import_str = ""
            base_str = ""

            for key, val in indices.items():
                import_str += f"from .{val} import {key}\n"
                base_str += f"{key}, "
            f.write(import_str + "\n\n")
            f.write(tmpl % base_str[:-2])

        print(f"{name} 驱动已创建。")

    @classmethod
    def enrich_parser(cls, sub_parser):
        super().enrich_parser(sub_parser)
        sub_parser.add_argument(
            "-b", "--base",
            default="apistellar.helper:RestfulApi",
            help="rpc基类 eg：apistellar.helper:RestfulApi")

        sub_parser.add_argument(
            "-ct", "--conn-timeout", type=int,
            help="连接超时时间， 单位：秒")

        sub_parser.add_argument(
            "-rt", "--read-timeout", type=int,
            help="数据读取时间， 单位：秒")



import os
import sys
import asyncio
import inspect

from apistar import Route
from apistellar.bases.manager import Manager
from apistellar.helper import get_real_method

from .mocker import Mocker


class ConsoleManager(Manager):
    """
    交互式调式命令 apistar-console
    """
    def __init__(self):
        self.initialize(os.getcwd())
        self.finalize()
        self.mock_keys = list()

    async def resolve(self, type):
        def wrapper(arg: type=None):
            return arg

        route = Route("/", "post", wrapper)
        state = dict(self.state)
        state["route"] = route
        return await self.injector.run_async([wrapper], state)

    def mock(self, mocker):
        """
        构建测试数据
        :param mocker: Mocker对象
        :return:
        """
        for type, val, param_name in mocker:
            param_name = param_name or type.__class__.__name__.lower()
            parameter = inspect.Parameter(
                param_name, inspect._POSITIONAL_OR_KEYWORD, annotation=type)
            for component in self.injector.components:
                if component.can_handle_parameter(parameter):
                    identity = component.identity(parameter)
                    break
            else:
                raise RuntimeError(f"Type: {type} cannot be mocked! ")

            self.mock_keys.append(identity)
            self.state[identity] = val
            self.injector.initial[identity] = type

    @staticmethod
    def _await(awaitable):
        """
        模拟await关键字
        :param awaitable:
        :return:
        """
        loop = asyncio.get_event_loop()
        task = loop.create_task(awaitable.__await__())
        loop.run_until_complete(task)
        return task.result()

    def clear(self):
        """
        清除mock数据
        :return:
        """
        for key in self.mock_keys:
            del self.state[key]
            del self.injector.initial[key]

    def start(self):
        if sys.version_info < (3, 7, 0):
            locals()["await"] = self._await
        else:
            locals()["_await"] = self._await
        mock = self.mock

        def inject(cls):
            return self._await(self.resolve(cls))

        from IPython.core import formatters
        # bugfix
        formatters.get_real_method = get_real_method

        # 延迟加载
        from IPython import embed

        embed()


def main():
    ConsoleManager().start()



import sys
import typing
import logging
import inspect
import traceback

from apistar import App, http
from flask.sessions import SecureCookieSessionInterface

from ..helper import HookReturn
from .entities import Session, DummyFlaskApp, Local


class SessionHook(object):

    def __init__(self):
        self.session_interface = SecureCookieSessionInterface()

    def on_response(self,
                    app: DummyFlaskApp,
                    resp: http.Response,
                    session: Session):
        if session is not None:
            self.session_interface.save_session(app, session, resp)


class ErrorHook(object):
    """
    处理异常增加响应码
    """
    errors = {999: "Unknown error"}
    fmt = '[{asctime}] {name} {levelname}: {message}'

    def __init__(self):
        self.logger = logging.getLogger("ErrorHook")
        self.logger.propagate = False
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(10)
        handler.setFormatter(logging.Formatter(self.fmt, style="{"))
        self.logger.addHandler(handler)

    def on_error(self, error: Exception, app: App) -> http.Response:
        """
        Handle error
        """
        args = list()
        if error.args:
            data = error.args[0]
            if not isinstance(data, (tuple, list)):
                if isinstance(data, int):
                    args.append(data)
                    args.append(None)
                else:
                    args.append(999)
                    args.append(data)
            else:
                args.extend(data)
        else:
            args.append(None)

        if not isinstance(args[0], int):
            args.insert(0, 999)

        args.extend([None, None, None])
        code, message, extra, *_ = args
        # apistar不支持在on_request时打断后续执行直接返回response
        # 所以在只能通过raise异常来通过异常参数传递响应。
        if isinstance(message, http.Response):
            return message

        if message is None:
            message = self.errors.get(code, "Not configured error")

        payload = {
            "type": "normal",
            "code": code,
            "errcode": getattr(message, "code", code),
            "message": message,
            "extra": extra,
        }
        detail = "".join(traceback.format_exc())
        if app.debug:
            payload["detail"] = detail
        self.logger.error(
            "Error happened: %s extra: %s trace:\n%s. ", message, extra, detail)
        return http.JSONResponse(payload)

    @classmethod
    def register(cls,
                 errors: typing.Union[typing.List[typing.Tuple[int, str]],
                                      typing.Mapping[int, str]]):
        cls.errors.update(errors)


class AccessLogHook(object):
    fmt = '{host} - - [{asctime}] {method} {path} {protocol}' \
          ' {status} {content_length} {agent}'

    def __init__(self):
        self.logger = logging.getLogger("access")
        self.logger.propagate = False
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(10)
        handler.setFormatter(logging.Formatter(self.fmt, style="{"))
        self.logger.addHandler(handler)

    def on_response(self,
                    host: http.Host,
                    path: http.Path,
                    protocol: http.Scheme,
                    method: http.Method,
                    resp: http.Response,
                    user_agent: http.Header,
                    string: http.QueryString):
        if string:
            path = path + "?" + string
        self.log(host, path, protocol, method, resp.status_code,
                 resp.headers["Content-Length"], user_agent)
        return resp

    on_error = on_response

    def log(self, host, path, protocol, method, status, content_length, agent):
        self.logger.info("", extra={"host": host,
                                    "path": path,
                                    "protocol": protocol,
                                    "method": method,
                                    "status": status,
                                    "content_length": content_length,
                                    "agent": agent})


class WebContextHook(Local):

    def __init__(self):
        params = [inspect.Parameter("self", inspect._POSITIONAL_OR_KEYWORD)]

        for param_name, param_type in self.local_variable.items():
            params.append(inspect.Parameter(
                param_name, inspect._POSITIONAL_OR_KEYWORD,
                annotation=param_type))

        self.on_request.__func__.__signature__ = inspect.Signature(params)

        self.on_request.__func__.__annotations__.update(self.local_variable)

    def __getattr__(self, item):
        raise AttributeError()

    def on_request(self, **kwargs):
        self.coroutinelocal[self.current_task()] = kwargs


class Hook(object):
    """
    Hook基类，继承自此基类的hook可以自动发现。
    """
    order = 1


def Return(return_value):
    if isinstance(return_value, str):
        return_value = http.HTMLResponse(return_value)
    elif not isinstance(return_value, http.Response):
        return_value = http.JSONResponse(return_value)
    raise HookReturn(return_value)


import os
import asyncio
import logging
import traceback

from apistar import ASyncApp, exceptions
from apistar.http import Response, JSONResponse
from apistar.server.components import ReturnValue
from apistar.server.asgi import ASGIScope, ASGISend

from apistellar.bases.entities import settings
from apistellar.bases.websocket import WebSocketApp
from apistellar.document import ShowLogPainter, AppLogPainter
from apistellar.bases.components import Component, ComposeTypeComponent
from apistellar.bases.hooks import WebContextHook, ErrorHook, \
    AccessLogHook, SessionHook, Hook
from apistellar.helper import TypeEncoder, find_children, enhance_response

__all__ = ["Application"]
enhance_response(Response)
# 可以序列化Type子类
JSONResponse.options["default"] = TypeEncoder().default

del JSONResponse.charset


RESP_BUFFER_SIZE = 1024000


class FixedAsyncApp(ASyncApp):

    def exception_handler(self, exc: Exception) -> Response:
        if isinstance(exc, exceptions.HTTPException):
            payload = {
                "type": "http",
                "code": exc.status_code,
                "errcode": exc.status_code,
                "message": exc.detail,
            }
            if self.debug:
                payload["detail"] = "".join(traceback.format_exc())
            return JSONResponse(payload, exc.status_code, exc.get_headers())
        raise exc

    def error_handler(self, return_value: ReturnValue) -> Response:
        if isinstance(return_value, Response):
            return return_value
        return super().error_handler()

    async def read(self, response):
        coroutine = response.content.read(RESP_BUFFER_SIZE)
        if asyncio.iscoroutine(coroutine) or asyncio.isfuture(coroutine):
            return await coroutine
        else:
            return coroutine

    async def finalize_asgi(self,
                            response: Response,
                            send: ASGISend,
                            scope: ASGIScope):
        if response.exc_info is not None:
            if self.debug or scope.get('raise_exceptions', False):
                exc_info = response.exc_info
                raise exc_info[0].with_traceback(exc_info[1], exc_info[2])

        await send({
            'type': 'http.response.start',
            'status': response.status_code,
            'headers': [
                [key.encode(), value.encode()]
                for key, value in response.headers
            ]
        })
        if hasattr(response.content, "read"):
            body = await self.read(response)
            while body:
                await send({
                    'type': 'http.response.body',
                    'body': body,
                    "more_body": True,
                })
                body = await self.read(response)
        else:
            body = response.content
        await send({
            'type': 'http.response.body',
            'body': body
        })

    def __call__(self, scope):
        if scope["type"] != "websocket":
            return super(FixedAsyncApp, self).__call__(scope)
        else:
            return WebSocketApp(scope, self)


def application(app_name,
                packages=None,
                static_url='/static/',
                settings_path="settings",
                debug=True,
                current_dir="."):
    """
       可以动态发现当前项目根目录下所有controller中的handler
    """
    logger = logging.getLogger(app_name)
    os.chdir(current_dir)

    with AppLogPainter(logger.debug, current_dir, settings_path).paint() as routes:
        components = find_children(Component)
        # ComposeTypeComponent是用来兜底的，所以要放到最后
        components.append(ComposeTypeComponent())
        custom_hooks = sorted(find_children(Hook), key=lambda x: x.order)
        hooks = [WebContextHook(), AccessLogHook(),
                 SessionHook(), ErrorHook()] + custom_hooks
        app = FixedAsyncApp(
            routes,
            template_dir=settings.get("TEMPLATE_DIR"),
            static_dir=settings.get("STATIC_DIR"),
            packages=packages,
            schema_url=None,
            docs_url=None,
            static_url=static_url,
            components=components,
            event_hooks=hooks)
        app.debug = debug
        return app


Application = application


def show_routes():
    formatter = "{:<40} {:<7} {:<40} {:<}"

    def show_format(method, parttern, name, ca_name):
        return formatter.format(name, method, parttern, ca_name)

    with ShowLogPainter(show_format).paint() as routes:
        if routes:
            print(formatter.format(
                "Name", "Method", "URI Pattern", "Controller#Action"))
